﻿# Copyright (c) 2013 SAS Institute Inc. 
# Unpublished - All Rights Reserved 
#  
# emi_include.ps1
# This file contains common variable and function definitions for use by
# SAS Environment Manager plugin scripts written by the EMI team.
 
# USAGE
# This script is intended to be sourced into other scripts and not run by itself
# Several variables are defined, but if already defined the existing value is used.
# This allows default behavior with an override mechanism.
# PowerShell checks variables first in the should be set as environment variables,
# in the global scope

# The following variables are defined and can be overridden (* indicates the default)
#    _LOG_DIR=[directory containing log files] 
#    _LOG_FILENAME=[_LOG_DIR/file name]
#    _LOG_LEVEL=[TRACE|DEBUG|INFO|WARN|ERROR|*OFF]
#    _LOG_FILENAME_USE_TIMESTAMP=[true|false*]
#    _LOG_FILENAME_USE_DAYROLLOVER=[true|false*]

# The following variables are defined by sourcing this script
# ---------------------------------------
# General Variables
# ---------------------------------------
# _SCRIPT_FILENAME - Identifies the absolute path to the calling script that sourced this file
# _SCRIPT_PATH     - Identifies the absolute path to the parent directory of the script
#                    that sourced this file
# _TIME_STR        - Identifies the time of execution of the script to the second
#                  - (e.g. 20130318_134510 = 3/18/13 1:45:10pm)
#
# ---------------------------------------
# SAS Environment Manager Agent Variables
# ---------------------------------------
#   _PLUGIN_NAME     - Name of the plugin that deployed this script
#   _AGENT_DIRECTORY - Base install directory of the SAS Environment Manager Agent
#                      (e.g. /install/cfgsas1/config/Lev1/Web/SASEnvironmentManager/agent-5.0.0-EE)
# The values of these variables will be empty ("") if this script is sourced outside
# of a plugin's script work directory


#region Log functions
$LogDay = (Get-Date).DayOfWeek

function LogObject() {
    param([string] $comment,
	      [object] $obj)
	if ((test-path $_LOG_DIR) -eq $false) {
		mkdir -p "$_LOG_DIR" >$null
	    }
	if ($_LOG_FILENAME_USE_TIMESTAMP -eq $false) {				
		if ($_LOG_FILENAME_USE_DAYROLLOVER -eq $true) {
			if ($LogDay -ne (Get-Date).DayOfWeek) {	
				$daylog = $LogFileBaseName + "_" + $LogDay + ".log"
				if (Test-Path $daylog) {
					Remove-Item $daylog
					}
				Move-Item $LOG_FILENAME $daylog
				$LogDay = (Get-Date).DayOfWeek
				}
			}
		}
	$_LOG_FILE = "$_LOG_DIR\$_LOG_FILENAME"
    $_LOG_WRITABLE = $true    
    if (test-path $_LOG_FILE) {
        $file = Get-Item $_LOG_FILE
        if ($file.attributes -band [system.IO.FileAttributes]::ReadOnly) {
            $_LOG_WRITABLE = $false
            }
        }
    
    if (($_LOG_WRITABLE) -and (($_LOG_FILE).Length -gt 0)) {
        $tstamp = (get-date).ToString('yyyy-MM-d H:mm:ss.fffff K')
		if ($obj -eq $null) {
        	"[$tstamp] $comment" >> "$_LOG_FILE"
			}
		else {
        	"[$tstamp] ----" >> "$_LOG_FILE"
        	"$comment" >> "$_LOG_FILE"			
        	write-output $obj >> "$_LOG_FILE"
        	"----" >> "$_LOG_FILE"			
			}
        }
    elseif ($_WARNED_UNABLE_TO_WRITE_LOG) {
        "WARNING: Unable to write log file [$_LOG_FILE]"
        $_WARNED_UNABLE_TO_WRITE_LOG = $true
        }
    }

# LogObjectDebug()
#     Log debug object
#     The object is only logged if the _LOG_LEVEL is set to _LOG_LEVEL_DEBUG or higher 
function LogObjectDebug() {
    param([string] $comment,
		  [object] $obj)
    if ($_LOG_LEVEL_INT -ge $_LOG_LEVEL_DEBUG) {
        LogObject $comment $obj
        }
    }

# Log()
#     Appends the provided message to the end of the configured log file.
#     This function ignores the configured _LOG_LEVEL and always logs the message. 
# Pre-Condition:
#     _LOG_DIR has been set to an existing directory
#     _LOG_FILENAME has been set and the file is writable  
# Parameters:
#     $1 - The message to append to the log file
function Log() {
    param([string] $msg)
	LogObject $msg $null
    }

# LogDebug()
#     Log debug message (usually not enabled by default)
#     The message is only logged if the _LOG_LEVEL is set to _LOG_LEVEL_DEBUG or higher 
function LogDebug() {
    param([switch] $toTerm,
          [string] $msg)
    if ($_LOG_LEVEL_INT -ge $_LOG_LEVEL_DEBUG) {
        Log "[DBUG] $msg"
        if ($toTerm) {
            write-host "[DEBUG] $msg"
            }
#        Write-Debug $msg
        }
    }

# LogInfo()
#     Log info message (usually enabled by default)
#     The message is only logged if the _LOG_LEVEL is set to _LOG_LEVEL_INFO or higher 
function LogInfo() {
    param([switch] $toTerm,
          [string] $msg)
    if ($_LOG_LEVEL_INT -ge $_LOG_LEVEL_INFO) {
        Log "[INFO] $msg"
        if ($toTerm) {
            write-host "[INFO] $msg"
            }
#        Write-Verbose $msg
        } 
    }

# LogWarn()
#     Log warning message
#     The message is only logged if the _LOG_LEVEL is set to _LOG_LEVEL_WARN or higher 
function LogWarn() {
    param([switch] $toTerm,
          [string] $msg)
    if ($_LOG_LEVEL_INT -ge $_LOG_LEVEL_WARN) {
        Log "[WARN] $msg"
        if ($toTerm) {
            write-host "[WARNING] $msg"
            }
#        Write-Warning $msg
        } 
    }

# LogError()
#     Log error message
#     The message is only logged if the _LOG_LEVEL is set to _LOG_LEVEL_ERROR or higher 
function LogError() {
    param([switch] $toTerm,
          [string] $msg)
    Log "[ERR_] $msg"
    if ($toTerm) {
        write-host "[ERROR] $msg"
        }
    if ($_LOG_LEVEL_INT -ge $_LOG_LEVEL_DEBUG) {
		Log "Stacktrace`r`n" + (Get-PSCallStack | format-list command, location, arguments)
		}
#    Write-Error $msg
    }

# LogTrace()
#     Log trace message
#     The message is only logged if the _LOG_LEVEL is set to _LOG_LEVEL_TRACE or higher 
function LogTrace() {
    param([switch] $toTerm,
          [string] $msg)
    if ($_LOG_LEVEL_INT -ge $_LOG_LEVEL_TRACE) {
        Log "[TRC_] $msg"
        if ($toTerm) {
            write-host "[TRACE] $msg"
            }
#        Write-Verbose $msg
        } 
    }

# LogTerm()
#     Log to terminal and an info message to the log.
function LogTerm() {
    param([string] $msg)
    LogInfo -toTerm $msg
    }

# LogSysError()
#     Log to error stream and an error message to the log.
function LogSysError() {
    param([string] $msg)
    LogError -toTerm "$msg"
    # should I be writing to the event log here? 
    }

# LogTermWarn()
#     Log to terminal and a warning message to the log.
function LogTermWarn() {
    param([string] $msg)
    LogWarn -toTerm "$msg"
    }

# LogTermError()
#     Log to terminal and an error message to the log.
function LogTermError() {
    param([string] $msg)
    LogError -toTerm "$msg"
    }
#endregion Log functions


# ====================================================================
# VARIABLES
# ====================================================================

$_EMI_INCLUDE_NAME     = "SAS EMI Team Common Functions"
$_EMI_INCLUDE_VERSION  = "00.00.00"
$_EMI_INCLUDE_BUILD_ID = "201200000000"

# Get a string of yyyymmddhhmmss to use for identifying time of execution
$_TIME_STR = (get-date).ToString("yyyyMMddhhmmss")

# get script path and name
# try to get the script that invoked us
$pth = $MyInvocation.ScriptName
# if scriptname isn't set try parsing the position message
if ($pth.Length -eq 0) {
	$pth = $MyInvocation.PositionMessage
	$ix = $pth.IndexOf("At ")
	if ($ix -gt -1) {
		$pth = $pth.SubString($ix + "At ".Length)
		$ix = $pth.LastIndexOf(".ps1:")
		if ($ix -gt -1) {		
			$pth = $pth.SubString(0, $ix + ".ps1".Length)
			}
		}
	}
	
# $MyInvocation.MyCommand (System.Management.Automation.ScriptInfo)
if ($pth.Length -eq 0) {
	$pth = $MyInvocation.MyCommand.Path
	if ($pth.Length -eq 0) {	
		$pth = $MyInvocation.MyCommand.Definition	
		if ($pth.Length -eq 0) {
			# try to use current working dir
			$pth = Join-Path Get-Location "emi_include.ps1"
			}
		}
	}

$_SCRIPT_PATH      = split-path -parent $pth # path to script	
$_SCRIPT_FILENAME  = split-path -leaf $pth # name of script, no path

# Set Agent Variables
if ((test-path variable:\_PLUGIN_NAME) -eq $false) {
	if (test-path env:_PLUGIN_NAME) {
	    $_PLUGIN_NAME = $env:_PLUGIN_NAME
	    }
	else {
	    $_PLUGIN_NAME = (Get-Item $_SCRIPT_PATH).BaseName # this should end up looking like "emi_include" or some such
	    }
	}
if ($_PLUGIN_NAME.Length -eq 0) {
	$_PLUGIN_NAME = "Unknown"
	}

if ((test-path variable:\_AGENT_DIRECTORY) -eq $false) {
	if (test-path env:_AGENT_DIRECTORY) {
	    $_AGENT_DIRECTORY = $env:_AGENT_DIRECTORY
	    }
	else {
		$_AGENT_DIRECTORY = $_SCRIPT_PATH -replace '^(.+agent-[0-9]+\.[0-9]+\.[0-9]+-EE\\)(.)+$', "`$1"
		}
	}

# Configure logging
$_LOG_LEVEL_TRACE = 5
$_LOG_LEVEL_DEBUG = 4
$_LOG_LEVEL_INFO  = 3
$_LOG_LEVEL_WARN  = 2
$_LOG_LEVEL_ERROR = 1
$_LOG_LEVEL_OFF   = 0

if ((test-path variable:\_LOG_LEVEL) -eq $false) {
	if (test-path env:_LOG_LEVEL) {
    	$_LOG_LEVEL = $env:_LOG_LEVEL
    	}
	else {
	    $_LOG_LEVEL = "OFF"
	    }
	}
$_LOG_LEVEL_INT = Invoke-Expression "`$_LOG_LEVEL_$($_LOG_LEVEL)" 

if ((test-path variable:\_LOG_DIR) -eq $false) {
	if (test-path env:_LOG_DIR) {
	    $_LOG_DIR = $env:_LOG_DIR
	    }
	else {
		$_LOG_DIR = ""
		}
	}

if (($_LOG_DIR).Length -eq 0) {
    # Default location is agent log directory
    if (($_AGENT_DIRECTORY).Length -gt 0) {
	    $_LOG_DIR = Join-Path $_AGENT_DIRECTORY "log"
        }	    
	elseif (($_SCRIPT_PATH).Length -gt 0) {
	    $_LOG_DIR = "$_SCRIPT_PATH"	
        }
	else {
	    $_LOG_DIR = "."
        }
    }

if ((test-path variable:\_LOG_FILENAME) -eq $false) {
	if (test-path env:_LOG_FILENAME) {
	    $_LOG_FILENAME = $env:_LOG_FILENAME
	    }
	else {
	    $_LOG_FILENAME = ""
	    }
	}

if (($_LOG_FILENAME).Length -eq 0) {
    # Default filename is [_PLUGIN_NAME]_scripts.log
    if (($_PLUGIN_NAME).Length -ne 0) {
    	$_LOG_FILENAME = "$_PLUGIN_NAME" + "_scripts"
        }
	elseif (($_SCRIPT_FILENAME).Length -gt 0) {
	    $_LOG_FILENAME = ($_SCRIPT_FILENAME)
        }
    else {
        # Could not determine parent script filename
        $_LOG_FILENAME = "EMI"
        }
    }
	
if ($_LOG_FILENAME.EndsWith(".log")) {
	$LogFileBaseName = $_LOG_FILENAME.SubString(0, $_LOG_FILENAME.Length - 4)
	}
else {	
	$LogFileBaseName = $_LOG_FILENAME
	}	
	
if ((test-path variable:\_LOG_FILENAME_USE_TIMESTAMP) -eq $false) {
	if (test-path env:_LOG_FILENAME_USE_TIMESTAMP) {
		[System.Boolean] $b = $false
		[System.Boolean].TryParse($env:_LOG_FILENAME_USE_TIMESTAMP, [ref] $b)		
		$_LOG_FILENAME_USE_TIMESTAMP = $b
		}
	else {
		$_LOG_FILENAME_USE_TIMESTAMP = $false
		}
	}
if ($_LOG_FILENAME_USE_TIMESTAMP.GetType().FullName -ne "System.Boolean") {
	$_LOG_FILENAME_USE_TIMESTAMP = $false # redefine
	}

if ((test-path variable:\_LOG_FILENAME_USE_DAYROLLOVER) -eq $false) {
	if (test-path env:_LOG_FILENAME_USE_DAYROLLOVER) {
		[System.Boolean] $b = $false
		[System.Boolean].TryParse($env:_LOG_FILENAME_USE_DAYROLLOVER, [ref] $b)		
		$_LOG_FILENAME_USE_DAYROLLOVER = $b
		}
	else {
		$_LOG_FILENAME_USE_DAYROLLOVER = $false
		}
	}
if ($_LOG_FILENAME_USE_DAYROLLOVER.GetType().FullName -ne "System.Boolean") {
	$_LOG_FILENAME_USE_DAYROLLOVER = $false # redefine
	}

if ($_LOG_FILENAME_USE_TIMESTAMP -eq $true) {
	$LogFileBaseName = $LogFileBaseName + "_" + $_TIME_STR
	}

$_LOG_FILENAME = $LogFileBaseName + ".log"

LogDebug "_EMI_INCLUDE_NAME: $_EMI_INCLUDE_NAME"
LogDebug "_EMI_INCLUDE_VERSION: $_EMI_INCLUDE_VERSION"
LogDebug "_EMI_INCLUDE_BUILD_ID: $_EMI_INCLUDE_BUILD_ID"
LogObjectDebug "Variables" (Get-Variable)
LogObjectDebug "Environment" (Get-ChildItem env:*)
LogObjectDebug "$MyInvocation" $MyInvocation
LogDebug "_LOG_DIR = `"$_LOG_DIR`""
LogDebug "_LOG_FILENAME = `"$_LOG_FILENAME`""
LogDebug "_LOG_LEVEL = `"$_LOG_LEVEL`""
LogDebug "_LOG_FILENAME_USE_TIMESTAMP = `"$_LOG_FILENAME_USE_TIMESTAMP`""
LogDebug "_SCRIPT_FILENAME = `"$_SCRIPT_FILENAME`""
LogDebug "_SCRIPT_PATH = `"$_SCRIPT_PATH`""
LogDebug "_TIME_STR = `"$_TIME_STR`""
LogDebug "_PLUGIN_NAME = `"$_PLUGIN_NAME`""
LogDebug "_AGENT_DIRECTORY = `"$_AGENT_DIRECTORY`""

LogInfo "emi_include.ps1 (End)"	