﻿# Copyright (c) 2013 SAS Institute Inc. 
# Unpublished - All Rights Reserved 
#
# pg_ha.sh
# This script verifies and/or reestablishs postgres master slave HA 
#
# USAGE
# Usage: 
#    pg_ha.sh [ -h | -help ]
#
#    pg_ha.sh [ -hostname hostname ]
#             [ -i | -install basepath ]
#             [ -l | -log logname ] 
#             [ -p | -port port ] 
#             [ -t | -trigger triggerfile ] 
#             [ -u | -user user ] 
#             [ -v | -verbose ] 
#             status | promote     
#
# RETURN CODES
# 0   : Success
# 2   : Failure  -  Needed for agent red failure condition
# ====================================================================
# Revision history
# ====================================================================
# Date        Developer   Change
# ----        ---------   ------
# 2013-04-12  grgran   	  Created

param (
    [alias("h")]
    [switch] 
    $help = $false,

    [string] 
    $hostname = "localhost",

    [alias("i")]
    [string] 
    $install = "",

    [alias("l")]
    [string] 
    $log = "default",

    [alias("p")]
    [uint16] 
    $port = 5432,

    [alias("t")]
    [string] 
    $trigger = "",

    [alias("u")]
    [string] 
    $user = "postgres",

    [switch] 
    $v = $false,

    [Parameter(Position = 0)] # this causes errorlevel to be 0 in calling bat file
	                          # adding $Host.SetShouldExit before exits to address this
    [ValidateSet("status", "promote")]
    [string] $choice
    )

$ActionPreferenceNS = [System.Management.Automation.ActionPreference] 

if ($log -eq "default") {
    $env:_LOG_FILENAME = "pgha_scripts.log"
    }
elseif ($log.Length -ne 0) {
    $env:_LOG_FILENAME = split-path -Leaf $log
    $env:_LOG_DIR      = split-path -Parent $log
    }

if ($v) {
    $verbose = $v
    }
elseif ($VerbosePreference -ne $ActionPreferenceNS::SilentlyContinue) {
    $verbose = $true
    }

if ($verbose) {
    $env:_LOG_LEVEL="DEBUG"
    }

# Include common functions
$ScriptPath = $MyInvocation.ScriptName
if ($ScriptPath.Length -eq 0) {
	$ScriptPath = $MyInvocation.InvocationName
    }
if ($ScriptPath.Length -eq 0) {
    $ScriptPath = "."
    }
else {    
    $ScriptPath = split-path -parent $ScriptPath
    }

. "$ScriptPath\emi_include.ps1"
    
$SCRIPT_NAME = "SAS Postgres HA script"
$SCRIPT_VERSION = "9.4.0.20130322"
$SCRIPT_BUILD_ID = "201303221500"

LogInfo "SCRIPT_NAME: $SCRIPT_NAME"
LogDebug "SCRIPT_VERSION: $SCRIPT_VERSION"
LogDebug "SCRIPT_BUILD_ID: $SCRIPT_BUILD_ID"

Log "* START *"
LogDebug "Arguments----------"
LogDebug "help     = $help"
LogDebug "hostname = $hostname"
LogDebug "install  = $install"
LogDebug "log      = $log"
LogDebug "port     = $port"
LogDebug "user     = $user"
LogDebug "v        = $v"
LogDebug "choice   = $choice"
LogDebug "-------------------"    


# ====================================================================
# FUNCTIONS
# ====================================================================
function ShowUsage() {
@"
This script verifies and/or reestablishs postgres master slave HA 

USAGE
Usage: 
   pg_ha.sh [ -h | -help ]

   pg_ha.sh [ -hostname hostname ]
            [ -l | -log logname ] 
            [ -i | -install binpath ]
            [ -p | -port port ] 
            [ -t | -trigger triggerfile ] 			
            [ -u | -user user ] 
            [ -v | -verbose ] 
            status | promote     

RETURN CODES
0   : Success
2   : Failure  -  Needed for agent red failure condition
"@
    }

function GetBinPath() {
	LogDebug "GetBinPath() entered"  
	
    $rv = New-Object object
    $binPath = ""
    $errorMsg = ""
    $returnCode = 0
    add-member -InputObject $rv -MemberType NoteProperty -Name BinPath $binPath
    add-member -InputObject $rv -MemberType NoteProperty -Name ErrorMsg $errorMsg
    add-member -InputObject $rv -MemberType NoteProperty -Name ReturnCode $returnCode
    
    # look in the registry to see if you can find the path to psql.exe
    $eap = $ErrorActionPreference
    $ErrorActionPreference = "Stop"
    $pgRegSrvKey = "hklm:\SYSTEM\CurrentControlSet\Services"
    $pgRegKey = "$pgRegSrvKey\post*"
    try {
        $name = Get-ChildItem $pgRegKey | select -ExpandProperty Name | Split-Path -Leaf
		LogDebug "Postgres versions in registry: $name"	
        }
    catch [System.Exception] {
        $rv.ErrorMsg = "Can't find Postgres registry service key" + [Environment]::NewLine
        $rv.ErrorMsg += $_.Message
        $rv.ReturnCode = 2
		LogError $rv.ErrorMsg		
        $rv
        return
        }
    # expecting name to contain something like this
    #    postgresql-x64-9.1
    $pgkey = $name | where {$_.EndsWith("9.1")}
    if ([System.String]::IsNullOrEmpty($pgkey)) {
        $pgkey = $name | select -First
        }
        
    try {
        $p = Get-ItemProperty "$pgRegSrvKey\$pgkey" | select -ExpandProperty ImagePath
        # $p should now look something like this:
        # C:/Program Files/PostgreSQL/9.1/bin/pg_ctl.exe runservice -N "postgresql-x64-9.1" -D "C:/Program Files/PostgreSQL/9.1/data" -w
		LogDebug "$pgRegSrvKey\$pgkey: `"$p`""			
        }
    catch [System.Exception] {    
        $rv.ErrorMsg = "Can't find Postgres ImagePath" + [Environment]::NewLine
        $rv.ErrorMsg += $_.Message
        $rv.ReturnCode = 2
		LogError $rv.ErrorMsg
        $rv
        return
        }
        
    try {        
        $ix = $p.IndexOf("pg_ctl.exe ")
        if ($ix -le -1) {
            $rv.ErrorMsg = "Can't find Postgres binary path"
            $rv.ReturnCode = 2
			LogError $rv.ErrorMsg			
            $rv
            return
            }
        $rv.BinPath = $p.Substring(0, $ix)
		LogDebug "Found BinPath: `"$rv.BinPath`""
        }
    catch [System.Exception] {    
        $rv.ErrorMsg = "Non-specific error" + [Environment]::NewLine
        $rv.ErrorMsg += $_.Message
        $rv.ReturnCode = 2
		LogError $rv.ErrorMsg		
        $rv
        return
        }
    $ErrorActionPreference = $eap
	LogDebug "GetBinPath() exited" 	
    $rv
    }

function ValidateBinPath() {
    param ([string] $binPath)
	LogDebug "ValidateBinPath() entered"
	LogDebug "BinPath: '$binPath'"  
	
    $script:psqlCmd = Join-Path "$binPath" "psql.exe"
	LogDebug "Validating psql command: `"$script:psqlCmd`""
    if ((test-path $script:psqlCmd -pathType leaf) -eq $false) {
        LogError "Can't find Postgres psql.exe file"
		$Host.SetShouldExit(2);
        exit 2
        }
    }

function SubmitSql {
    param([string] $command) 
    # setup args (quoting can be tricky)
    $cmdargs = @("-U", "`"$user`"",
                 "-d", "postgres",
                 "-h", "`"$hostname`"",
                 "-p", "$port",
                 "-w",
                 "-qAtX",
                 "-c", "`"$command`"")
	LogDebug "SubmitSql(): `"$script:psqlCmd $cmdargs`""
    $result = & $script:psqlCmd $cmdargs
    $rcode = $LASTEXITCODE
	LogDebug "SubmitSql() Result: `"$result`""	
	LogDebug "SubmitSql() ReturnCode : `"$rcode`""		
    $rv = New-Object object
    add-member -InputObject $rv -MemberType NoteProperty -Name Result $result
    add-member -InputObject $rv -MemberType NoteProperty -Name ReturnCode $rcode
    $rv
    }

function ParseXlogLocation {
    param ([string] $text)
	LogDebug "ParseXlogLocation($text)"	
    $text = $text.Replace("/", "")
    $tv = ""
    if ($parts.Length -gt 0) {
        $tv = [Convert]::ToInt64($text, 16).ToString()
        }
    $rv = New-Object object
    add-member -InputObject $rv -MemberType NoteProperty -Name TestValue $tv
	LogDebug "ParseXlogLocation($text) => $tv"
    $rv        
    }
    
# ====================================================================
# Main
# ====================================================================
if ($help) {
    ShowUsage
	LogDebug "Help options used exiting after ShowUsage"
	$Host.SetShouldExit(2);	
    exit 2 # this will set $LASTEXITCODE = 2
    }

$pgpaths = New-Object object
if ($install.Length -eq 0) {
    $pgpaths = GetBinPath
	}
else {
    add-member -InputObject $pgpaths -MemberType NoteProperty -Name BinPath ($install + "\bin")
    add-member -InputObject $pgpaths -MemberType NoteProperty -Name ErrorMsg ""
    add-member -InputObject $pgpaths -MemberType NoteProperty -Name ReturnCode 0
    }

if ($choice.Length -eq 0) {
    $PG_CHOICE = "status"
    }
else {
    $PG_CHOICE = $choice
    }

if ($user.Length -ne 0) {
    $PG_USER = $user
    }

if ($PG_USER.Length -eq 0) {
    $PG_USER = "postgres"
    }

ValidateBinPath $pgpaths.BinPath

LogDebug "PG_CHOICE: `"$PG_CHOICE`""
LogDebug "PG_USER: `"$PG_USER`""
LogDebug "BinPath: `"$($pgpaths.BinPath)`""
#-----------    

    
$ccon = SubmitSql "SELECT COUNT(*) from pg_stat_activity;"
if (($ccon.Result.Length -eq 0) -or ($ccon.ReturnCode -ne 0)) {
	LogTermError ("Database exist test failed, curr_conn=" + $ccon.Result +
				  ", Bad rc=" + $ccon.ReturnCode)
	$Host.SetShouldExit(2);				  
	exit 2
    }
LogInfo "$PG_USER SELECT COUNT(*) from pg_stat_activity result=$($ccon.Result) rc=$($mcon.ReturnCode)"
#-----------


$mcon = SubmitSql "SHOW max_connections;"
if ($mcon.ReturnCode -ne 0) {
	LogTermError "Database max_conn query failed mcon=$($mcon.Result), Bad return code=$($mcon.ReturnCode)"
	$Host.SetShouldExit(2);	
	exit 2
    }
LogInfo "$PG_USER SHOW max_connections result=$($mcon.Result) rc=$($mcon.ReturnCode)"
#-----------


# compute a ratio of how "busy" the server is
$curCons = 0.0
$b = [Double]::TryParse($ccon.Result, [ref] $curCons)
$maxCons = 0.0
$b = [Double]::TryParse($mcon.Result, [ref] $maxCons)
if ($maxCons -ne 0.0) {
    $pctused = $curCons / $maxCons
    }
else {
    $pctused = [Double]::NaN
    }

LogInfo "Server is $pctused % busy"
#-----------


# So database exists, now test for slave LSN value on master via pg_current_xlog_location query 
# master or pg_last_xlog_receive_location on slave to determine if I am on master or slave
#
$PG_TYPE="Unknown"
$recover = SubmitSql "SELECT pg_is_in_recovery();"
if ($recover.Result -eq "f") {
	$PG_TYPE="master"
	$masterchk= SubmitSql "SELECT pg_current_xlog_location();"
	$parts = $masterchk.Result -split "/"
	if ($parts.Length -eq 2) {  
		$p1 = [Convert]::ToUInt32($parts[0], 16)
		$p2 = [Convert]::ToUInt32($parts[1], 16)
		$master = "{0:D10}{1:D10}" -f $p1, $p2
		}
	else {
		$master = ""		
		}
	LogInfo "Master pg_current_xlog_location=$($masterchk.Result)"
	}
elseif ($recover.Result -eq "t") {
	$PG_TYPE="slave"
	$slavechk = SubmitSql "SELECT pg_last_xlog_receive_location();"
	$parts = $slavechk.Result -split "/"
	if ($parts.Length -eq 2) {  
		$p1 = [Convert]::ToUInt32($parts[0], 16)
		$p2 = [Convert]::ToUInt32($parts[1], 16)
		$slave = "{0:D10}{1:D10}" -f $p1, $p2
		}
	else {
		$slave = ""		
		}
	LogInfo "Slave SELECT pg_last_xlog_receive_location=$($slavechk.Result)"	
	# the above returns something like '0/16B6D78'
	}
else {
	LogError "pg_current_xlog & pg_last_xlog_receive tests failed, Return code=$($recover.ReturnCode) value [$($recover.Result)]"
	$Host.SetShouldExit(2);	
	exit 2
	}
LogInfo "master: `"$master`""
LogInfo "slave:  `"$slave`""

if ($PG_CHOICE -eq "status") {
	if ($PG_TYPE -eq "master") {
		"pgcurrentxlog=" + $master
		}
	else {
		"pglastxlog=" + $slave
		}
    "current=" + $ccon.Result 
    "pctused=" + "{0:N2}" -f $pctused
    }
elseif ($PG_CHOICE -eq "promote") {
	LogInfo "Promoting to master, trigger file [$trigger];"
	if ($PG_TYPE -eq "master") {
		LogTermError "On $PG_TYPE, Cannot promote master to master"
		$Host.SetShouldExit(2)
		exit 2
		}
    if (("$PG_TYPE" -eq "slave") -and (test-path $trigger)) {
		LogTermError "Triggerfile [$trigger] exists before creation!"
		$Host.SetShouldExit(2);		
		exit 2		
		}
	"x" >$trigger
	if ((Test-Path $trigger) -eq $false) {
		LogTermError "Could not create [$trigger]"
		$Host.SetShouldExit(2);		
		exit 2
		}
	sleep 5
	LogDebug "Created triggerfile [$trigger]"
	
	# now should confirm status t
	$chk = SubmitSql "SELECT pg_is_in_recovery();"
	if ($chk.Result -ne "f") {
		LogTermError "Promote failed, psql pg_is_in_recovery() returns=[$($chk.Result)]"
		$Host.SetShouldExit(2);		
		exit 2
		}
	LogInfo "Master confirmed, Promote successful"		
    }
else {
    # should never get here, but included for completeness
    LogTermError "Invalid internal argument, should never get there PG_CHOICE=$PG_CHOICE"
    }
LogDebug "=========== END OF EXECUTION =============="
Log "*  END  *"
# SIG # Begin signature block
# MIIZLQYJKoZIhvcNAQcCoIIZHjCCGRoCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUE4XJY0CrYBwOwA9TudXFV0ko
# 1iqgghQlMIID7jCCA1egAwIBAgIQfpPr+3zGTlnqS5p31Ab8OzANBgkqhkiG9w0B
# AQUFADCBizELMAkGA1UEBhMCWkExFTATBgNVBAgTDFdlc3Rlcm4gQ2FwZTEUMBIG
# A1UEBxMLRHVyYmFudmlsbGUxDzANBgNVBAoTBlRoYXd0ZTEdMBsGA1UECxMUVGhh
# d3RlIENlcnRpZmljYXRpb24xHzAdBgNVBAMTFlRoYXd0ZSBUaW1lc3RhbXBpbmcg
# Q0EwHhcNMTIxMjIxMDAwMDAwWhcNMjAxMjMwMjM1OTU5WjBeMQswCQYDVQQGEwJV
# UzEdMBsGA1UEChMUU3ltYW50ZWMgQ29ycG9yYXRpb24xMDAuBgNVBAMTJ1N5bWFu
# dGVjIFRpbWUgU3RhbXBpbmcgU2VydmljZXMgQ0EgLSBHMjCCASIwDQYJKoZIhvcN
# AQEBBQADggEPADCCAQoCggEBALGss0lUS5ccEgrYJXmRIlcqb9y4JsRDc2vCvy5Q
# WvsUwnaOQwElQ7Sh4kX06Ld7w3TMIte0lAAC903tv7S3RCRrzV9FO9FEzkMScxeC
# i2m0K8uZHqxyGyZNcR+xMd37UWECU6aq9UksBXhFpS+JzueZ5/6M4lc/PcaS3Er4
# ezPkeQr78HWIQZz/xQNRmarXbJ+TaYdlKYOFwmAUxMjJOxTawIHwHw103pIiq8r3
# +3R8J+b3Sht/p8OeLa6K6qbmqicWfWH3mHERvOJQoUvlXfrlDqcsn6plINPYlujI
# fKVOSET/GeJEB5IL12iEgF1qeGRFzWBGflTBE3zFefHJwXECAwEAAaOB+jCB9zAd
# BgNVHQ4EFgQUX5r1blzMzHSa1N197z/b7EyALt0wMgYIKwYBBQUHAQEEJjAkMCIG
# CCsGAQUFBzABhhZodHRwOi8vb2NzcC50aGF3dGUuY29tMBIGA1UdEwEB/wQIMAYB
# Af8CAQAwPwYDVR0fBDgwNjA0oDKgMIYuaHR0cDovL2NybC50aGF3dGUuY29tL1Ro
# YXd0ZVRpbWVzdGFtcGluZ0NBLmNybDATBgNVHSUEDDAKBggrBgEFBQcDCDAOBgNV
# HQ8BAf8EBAMCAQYwKAYDVR0RBCEwH6QdMBsxGTAXBgNVBAMTEFRpbWVTdGFtcC0y
# MDQ4LTEwDQYJKoZIhvcNAQEFBQADgYEAAwmbj3nvf1kwqu9otfrjCR27T4IGXTdf
# plKfFo3qHJIJRG71betYfDDo+WmNI3MLEm9Hqa45EfgqsZuwGsOO61mWAK3ODE2y
# 0DGmCFwqevzieh1XTKhlGOl5QGIllm7HxzdqgyEIjkHq3dlXPx13SYcqFgZepjhq
# IhKjURmDfrYwggSjMIIDi6ADAgECAhAOz/Q4yP6/NW4E2GqYGxpQMA0GCSqGSIb3
# DQEBBQUAMF4xCzAJBgNVBAYTAlVTMR0wGwYDVQQKExRTeW1hbnRlYyBDb3Jwb3Jh
# dGlvbjEwMC4GA1UEAxMnU3ltYW50ZWMgVGltZSBTdGFtcGluZyBTZXJ2aWNlcyBD
# QSAtIEcyMB4XDTEyMTAxODAwMDAwMFoXDTIwMTIyOTIzNTk1OVowYjELMAkGA1UE
# BhMCVVMxHTAbBgNVBAoTFFN5bWFudGVjIENvcnBvcmF0aW9uMTQwMgYDVQQDEytT
# eW1hbnRlYyBUaW1lIFN0YW1waW5nIFNlcnZpY2VzIFNpZ25lciAtIEc0MIIBIjAN
# BgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAomMLOUS4uyOnREm7Dv+h8GEKU5Ow
# mNutLA9KxW7/hjxTVQ8VzgQ/K/2plpbZvmF5C1vJTIZ25eBDSyKV7sIrQ8Gf2Gi0
# jkBP7oU4uRHFI/JkWPAVMm9OV6GuiKQC1yoezUvh3WPVF4kyW7BemVqonShQDhfu
# ltthO0VRHc8SVguSR/yrrvZmPUescHLnkudfzRC5xINklBm9JYDh6NIipdC6Anqh
# d5NbZcPuF3S8QYYq3AhMjJKMkS2ed0QfaNaodHfbDlsyi1aLM73ZY8hJnTrFxeoz
# C9Lxoxv0i77Zs1eLO94Ep3oisiSuLsdwxb5OgyYI+wu9qU+ZCOEQKHKqzQIDAQAB
# o4IBVzCCAVMwDAYDVR0TAQH/BAIwADAWBgNVHSUBAf8EDDAKBggrBgEFBQcDCDAO
# BgNVHQ8BAf8EBAMCB4AwcwYIKwYBBQUHAQEEZzBlMCoGCCsGAQUFBzABhh5odHRw
# Oi8vdHMtb2NzcC53cy5zeW1hbnRlYy5jb20wNwYIKwYBBQUHMAKGK2h0dHA6Ly90
# cy1haWEud3Muc3ltYW50ZWMuY29tL3Rzcy1jYS1nMi5jZXIwPAYDVR0fBDUwMzAx
# oC+gLYYraHR0cDovL3RzLWNybC53cy5zeW1hbnRlYy5jb20vdHNzLWNhLWcyLmNy
# bDAoBgNVHREEITAfpB0wGzEZMBcGA1UEAxMQVGltZVN0YW1wLTIwNDgtMjAdBgNV
# HQ4EFgQURsZpow5KFB7VTNpSYxc/Xja8DeYwHwYDVR0jBBgwFoAUX5r1blzMzHSa
# 1N197z/b7EyALt0wDQYJKoZIhvcNAQEFBQADggEBAHg7tJEqAEzwj2IwN3ijhCcH
# bxiy3iXcoNSUA6qGTiWfmkADHN3O43nLIWgG2rYytG2/9CwmYzPkSWRtDebDZw73
# BaQ1bHyJFsbpst+y6d0gxnEPzZV03LZc3r03H0N45ni1zSgEIKOq8UvEiCmRDoDR
# EfzdXHZuT14ORUZBbg2w6jiasTraCXEQ/Bx5tIB7rGn0/Zy2DBYr8X9bCT2bW+IW
# yhOBbQAuOA2oKY8s4bL0WqkBrxWcLC9JG9siu8P+eJRRw4axgohd8D20UaF5Mysu
# e7ncIAkTcetqGVvP6KUwVyyJST+5z3/Jvz4iaGNTmr1pdKzFHTx/kuDDvBzYBHUw
# ggV6MIIEYqADAgECAhAWywnjJw055iMTAk0RqCj/MA0GCSqGSIb3DQEBBQUAMIG0
# MQswCQYDVQQGEwJVUzEXMBUGA1UEChMOVmVyaVNpZ24sIEluYy4xHzAdBgNVBAsT
# FlZlcmlTaWduIFRydXN0IE5ldHdvcmsxOzA5BgNVBAsTMlRlcm1zIG9mIHVzZSBh
# dCBodHRwczovL3d3dy52ZXJpc2lnbi5jb20vcnBhIChjKTEwMS4wLAYDVQQDEyVW
# ZXJpU2lnbiBDbGFzcyAzIENvZGUgU2lnbmluZyAyMDEwIENBMB4XDTEzMDUyMTAw
# MDAwMFoXDTE2MDUyMDIzNTk1OVowgb0xCzAJBgNVBAYTAlVTMRcwFQYDVQQIEw5O
# b3J0aCBDYXJvbGluYTENMAsGA1UEBxMEQ2FyeTEbMBkGA1UEChQSU0FTIEluc3Rp
# dHV0ZSBJbmMuMT4wPAYDVQQLEzVEaWdpdGFsIElEIENsYXNzIDMgLSBNaWNyb3Nv
# ZnQgU29mdHdhcmUgVmFsaWRhdGlvbiB2MjEMMAoGA1UECxQDUiZEMRswGQYDVQQD
# FBJTQVMgSW5zdGl0dXRlIEluYy4wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK
# AoIBAQCQX7SORtH7+uBiyR+D1GNHEDQn0pZnuh1JJgf/HTQlO7L2S9XbT4VoFdQ+
# AgKfNX/R4l/i6LG9ppexLN3+X2NgEoNtu8JazVEnSbL1z4zwIwlI2x4K2QWno/Lr
# SurExvbCQ4+oeeENphgUJw3HWZZbbPAk5rtmpBRKCZqTMu6+GSPZbfy6IAMgY9WI
# N+Xn6fmvDs0/dtVXxoMGcd9cqAsZCzS2ca9+wE/Tr/RnuAU7HX1/VyveeBXASgph
# JU5eQUpgI+DnYmeePtXNATgbNg5SQ9nKRcFk9He7Fa57UYCoZ3nKM8ieLpe5sErE
# yJlTJ509FPuEzBFG71HfV5aMkyiXAgMBAAGjggF7MIIBdzAJBgNVHRMEAjAAMA4G
# A1UdDwEB/wQEAwIHgDBABgNVHR8EOTA3MDWgM6Axhi9odHRwOi8vY3NjMy0yMDEw
# LWNybC52ZXJpc2lnbi5jb20vQ1NDMy0yMDEwLmNybDBEBgNVHSAEPTA7MDkGC2CG
# SAGG+EUBBxcDMCowKAYIKwYBBQUHAgEWHGh0dHBzOi8vd3d3LnZlcmlzaWduLmNv
# bS9ycGEwEwYDVR0lBAwwCgYIKwYBBQUHAwMwcQYIKwYBBQUHAQEEZTBjMCQGCCsG
# AQUFBzABhhhodHRwOi8vb2NzcC52ZXJpc2lnbi5jb20wOwYIKwYBBQUHMAKGL2h0
# dHA6Ly9jc2MzLTIwMTAtYWlhLnZlcmlzaWduLmNvbS9DU0MzLTIwMTAuY2VyMB8G
# A1UdIwQYMBaAFM+Zqep7JvRLyY6P1/AFJu/j0qedMBEGCWCGSAGG+EIBAQQEAwIE
# EDAWBgorBgEEAYI3AgEbBAgwBgEBAAEB/zANBgkqhkiG9w0BAQUFAAOCAQEA8uhB
# qqTTBMT6tJbk3u/3PsskX0nT7RHkEwZtz3F/+19UiYFligjltaBdJd6iXaW5p+1F
# U9i+8mRbUSY4RF3aC44SL6xd6dDfvfLI9jOaEbcJt2OCgLf1rnfdzT4nMh7gzdOo
# zj8RkgLvnghmPinxEswmPovefpjP2IhvTYaPUzZnS/MQKNeuOdO8JdvQQ+2H2CqQ
# ewnYPhRl8H2T6qHGU0diuiskLTrocgjI4d2K4yJ3VbyjsJj1oJhhCr2vQNHZeurG
# XWBv7v1P+WCBBuVK+2grX09eGlU2y2EaqBQSt8ApaLTFlG51P1cxjSp+JkWsxFxe
# Bt+hm9724CdneBHRBDCCBgowggTyoAMCAQICEFIA5aolVvwahu2WydRLM8cwDQYJ
# KoZIhvcNAQEFBQAwgcoxCzAJBgNVBAYTAlVTMRcwFQYDVQQKEw5WZXJpU2lnbiwg
# SW5jLjEfMB0GA1UECxMWVmVyaVNpZ24gVHJ1c3QgTmV0d29yazE6MDgGA1UECxMx
# KGMpIDIwMDYgVmVyaVNpZ24sIEluYy4gLSBGb3IgYXV0aG9yaXplZCB1c2Ugb25s
# eTFFMEMGA1UEAxM8VmVyaVNpZ24gQ2xhc3MgMyBQdWJsaWMgUHJpbWFyeSBDZXJ0
# aWZpY2F0aW9uIEF1dGhvcml0eSAtIEc1MB4XDTEwMDIwODAwMDAwMFoXDTIwMDIw
# NzIzNTk1OVowgbQxCzAJBgNVBAYTAlVTMRcwFQYDVQQKEw5WZXJpU2lnbiwgSW5j
# LjEfMB0GA1UECxMWVmVyaVNpZ24gVHJ1c3QgTmV0d29yazE7MDkGA1UECxMyVGVy
# bXMgb2YgdXNlIGF0IGh0dHBzOi8vd3d3LnZlcmlzaWduLmNvbS9ycGEgKGMpMTAx
# LjAsBgNVBAMTJVZlcmlTaWduIENsYXNzIDMgQ29kZSBTaWduaW5nIDIwMTAgQ0Ew
# ggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQD1I0tepdeKuzLp1Ff37+TH
# Jn6tGZj+qJ19lPY2axDXdYEwfwRof8srdR7NHQiM32mUpzejnHuA4Jnh7jdNX847
# FO6G1ND1JzW8JQs4p4xjnRejCKWrsPvNamKCTNUh2hvZ8eOEO4oqT4VbkAFPyad2
# EH8nA3y+rn59wd35BbwbSJxp58CkPDxBAD7fluXF5JRx1lUBxwAmSkA8taEmqQyn
# bYCOkCV7z78/HOsvlvrlh3fGtVayejtUMFMb32I0/x7R9FqTKIXlTBdOflv9pJOZ
# f9/N76R17+8V9kfn+Bly2C40Gqa0p0x+vbtPDD1X8TDWpjaO1oB21xkupc1+NC2J
# AgMBAAGjggH+MIIB+jASBgNVHRMBAf8ECDAGAQH/AgEAMHAGA1UdIARpMGcwZQYL
# YIZIAYb4RQEHFwMwVjAoBggrBgEFBQcCARYcaHR0cHM6Ly93d3cudmVyaXNpZ24u
# Y29tL2NwczAqBggrBgEFBQcCAjAeGhxodHRwczovL3d3dy52ZXJpc2lnbi5jb20v
# cnBhMA4GA1UdDwEB/wQEAwIBBjBtBggrBgEFBQcBDARhMF+hXaBbMFkwVzBVFglp
# bWFnZS9naWYwITAfMAcGBSsOAwIaBBSP5dMahqyNjmvDz4Bq1EgYLHsZLjAlFiNo
# dHRwOi8vbG9nby52ZXJpc2lnbi5jb20vdnNsb2dvLmdpZjA0BgNVHR8ELTArMCmg
# J6AlhiNodHRwOi8vY3JsLnZlcmlzaWduLmNvbS9wY2EzLWc1LmNybDA0BggrBgEF
# BQcBAQQoMCYwJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLnZlcmlzaWduLmNvbTAd
# BgNVHSUEFjAUBggrBgEFBQcDAgYIKwYBBQUHAwMwKAYDVR0RBCEwH6QdMBsxGTAX
# BgNVBAMTEFZlcmlTaWduTVBLSS0yLTgwHQYDVR0OBBYEFM+Zqep7JvRLyY6P1/AF
# Ju/j0qedMB8GA1UdIwQYMBaAFH/TZafC3ey78DAJ80M5+gKvMzEzMA0GCSqGSIb3
# DQEBBQUAA4IBAQBWIuY0pMRhy0i5Aa1WqGQP2YyRxLvMDOWteqAif99HOEotbNF/
# cRp87HCpsfBP5A8MU/oVXv50mEkkhYEmHJEUR7BMY4y7oTTUxkXoDYUmcwPQqYxk
# bdxxkuZFBWAVWVE5/FgUa/7UpO15awgMQXLnNyIGCb4j6T9Emh7pYZ3MsZBc/D3S
# jaxCPWU21LQ9QCiPmxDPIybMSyDLkB9djEw0yjzY5TfWb6UgvTTrJtmuDefFmveh
# tCGRM2+G6Fi7JXx0Dlj+dRtjP84xfJuPG5aexVN2hFucrZH6rO2Tul3IIVPCglNj
# rxINUIcRGz1UUpaKLJw9khoImgUux5OlSJHTMYIEcjCCBG4CAQEwgckwgbQxCzAJ
# BgNVBAYTAlVTMRcwFQYDVQQKEw5WZXJpU2lnbiwgSW5jLjEfMB0GA1UECxMWVmVy
# aVNpZ24gVHJ1c3QgTmV0d29yazE7MDkGA1UECxMyVGVybXMgb2YgdXNlIGF0IGh0
# dHBzOi8vd3d3LnZlcmlzaWduLmNvbS9ycGEgKGMpMTAxLjAsBgNVBAMTJVZlcmlT
# aWduIENsYXNzIDMgQ29kZSBTaWduaW5nIDIwMTAgQ0ECEBbLCeMnDTnmIxMCTRGo
# KP8wCQYFKw4DAhoFAKBwMBAGCisGAQQBgjcCAQwxAjAAMBkGCSqGSIb3DQEJAzEM
# BgorBgEEAYI3AgEEMBwGCisGAQQBgjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqG
# SIb3DQEJBDEWBBScSSXsgQmStC7OnrhuWvxHjO/hEjANBgkqhkiG9w0BAQEFAASC
# AQBJg7hS0BFHh+80QYZ3H+TfAFknIgCfSLb6BtdYNMXbx3fultozxiTui58OcKf0
# AXnmSrJXs/jSN1/OneYpxoDLPt/4l+P+H8CwQV9e+BRk1eomjbzZa7bc1kHqIP1u
# oFdSpGCdw4m48R9G0scOxgRNJ4j2W7uxhXCFO4a3hTVYfmlU90t27zpIhr1DpQnT
# UEmPh8/cwNTPKCwkEyQcmM0objIMkFnMe29651NkPqvAdHxtJ4u0A+spGZ2Ypjju
# 5wabE7culpErRM4zyZPd8KL39HGYkrktbCXlEsKYhUJtAInkKhNTKwHhlt3rxmUM
# qVlCOExDk/c9OfPmU/IOTuFpoYICCzCCAgcGCSqGSIb3DQEJBjGCAfgwggH0AgEB
# MHIwXjELMAkGA1UEBhMCVVMxHTAbBgNVBAoTFFN5bWFudGVjIENvcnBvcmF0aW9u
# MTAwLgYDVQQDEydTeW1hbnRlYyBUaW1lIFN0YW1waW5nIFNlcnZpY2VzIENBIC0g
# RzICEA7P9DjI/r81bgTYapgbGlAwCQYFKw4DAhoFAKBdMBgGCSqGSIb3DQEJAzEL
# BgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTEzMDUyODE0MDI0NVowIwYJKoZI
# hvcNAQkEMRYEFIZhvyoWixbUA2kfHuDqBDkCVI0IMA0GCSqGSIb3DQEBAQUABIIB
# ABHXiFo0aXmKs7tYY77BQYaA8ljVv+y0WNV5UWwXQZk8lIWsG6GW+zT5y6Nlkuxi
# lKDAXC8EeeHt+ZeWr7aDvXRUEg9EEjTNFGUaL2mqyoOzKQzGpQceprPwfaOngIv3
# 6MFSBRpYmsDiPQZQAXbNGk72INNv2Leh8f4YI4slhCqlYoGUx/jmi2A2+nHdM091
# xenmLVJoICr1XOlTN4cldxzmGpthIYY0wcQhC5d105/zJ/7+bwBNLveJ7iqLANoy
# W0wK3tXSTPX+KoueZVOJVYQzppcZ4C3uXqOk+hMu9WfIxhGHvwktHWgS0krN9W01
# 6u8gsxseuheeGsvPPFjy2dE=
# SIG # End signature block
